""" Dataverse default metadata dictionary. """
DATAVERSE_METADATA = {
  'datasetVersion': {
    'license': {
      'name': 'CC0 1.0',
      'uri': 'http://creativecommons.org/publicdomain/zero/1.0'},
    'metadataBlocks': {
      'citation': {
        'displayName': 'Citation Metadata',
        'fields': [
          {
            'typeName': 'title',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'Replication Data for: Title'},
          {
            'typeName': 'subtitle',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'Subtitle'},
          {
            'typeName': 'alternativeTitle',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'Alternative Title'},
          {
            'typeName': 'alternativeURL',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'http://AlternativeURL.org'},
          {
            'typeName': 'otherId',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'otherIdAgency': {
                  'typeName': 'otherIdAgency',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'OtherIDAgency1'
                },
                'otherIdValue': {
                  'typeName': 'otherIdValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'OtherIDIdentifier1'
                }
              },
              {
                'otherIdAgency': {
                  'typeName': 'otherIdAgency',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'OtherIDAgency2'
                },
                'otherIdValue': {
                  'typeName': 'otherIdValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'OtherIDIdentifier2'
                }
              }
            ]
          },
          {
            'typeName': 'author',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'authorName': {
                  'typeName': 'authorName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'LastAuthor1, FirstAuthor1'
                },
                'authorAffiliation': {
                  'typeName': 'authorAffiliation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'AuthorAffiliation1'
                },
                'authorIdentifierScheme': {
                  'typeName': 'authorIdentifierScheme',
                  'multiple': False,
                  'typeClass': 'controlledVocabulary',
                  'value': 'ORCID'
                },
                'authorIdentifier': {
                  'typeName': 'authorIdentifier',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'AuthorIdentifier1'
                }},
              {
                'authorName': {
                  'typeName': 'authorName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'LastAuthor2, FirstAuthor2'
                },
                'authorAffiliation': {
                  'typeName': 'authorAffiliation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'AuthorAffiliation2'
                },
                'authorIdentifierScheme': {
                  'typeName': 'authorIdentifierScheme',
                  'multiple': False,
                  'typeClass': 'controlledVocabulary',
                  'value': 'ISNI'
                },
                'authorIdentifier': {
                  'typeName': 'authorIdentifier',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'AuthorIdentifier2'
                }}]},
          {
            'typeName': 'datasetContact',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'datasetContactName': {
                  'typeName': 'datasetContactName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'LastContact1, FirstContact1'},
                'datasetContactAffiliation': {
                  'typeName': 'datasetContactAffiliation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'ContactAffiliation1'},
                'datasetContactEmail': {
                  'typeName': 'datasetContactEmail',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'ContactEmail1@mailinator.com'}},
              {
                'datasetContactName': {
                  'typeName': 'datasetContactName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'LastContact2, FirstContact2'},
                'datasetContactAffiliation': {
                  'typeName': 'datasetContactAffiliation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'ContactAffiliation2'},
                'datasetContactEmail': {
                  'typeName': 'datasetContactEmail',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'ContactEmail2@mailinator.com'}
              }]},
          {
            'typeName': 'dsDescription',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'dsDescriptionValue': {
                  'typeName': 'dsDescriptionValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'DescriptionText1'},
                'dsDescriptionDate': {
                  'typeName': 'dsDescriptionDate',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1000-01-01'}},
              {
                'dsDescriptionValue': {
                  'typeName': 'dsDescriptionValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'DescriptionText2'},
                'dsDescriptionDate': {
                  'typeName': 'dsDescriptionDate',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1000-02-02'}}]},
          {
            'typeName': 'subject',
            'multiple': True,
            'typeClass': 'controlledVocabulary',
            'value': [
              'Agricultural Sciences',
              'Business and Management',
              'Engineering',
              'Law'
            ]
          },
          {
            'typeName': 'keyword',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'keywordValue': {
                  'typeName': 'keywordValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'KeywordTerm1'
                },
                'keywordVocabulary': {
                  'typeName': 'keywordVocabulary',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'KeywordVocabulary1'
                },
                'keywordVocabularyURI': {
                  'typeName': 'keywordVocabularyURI',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://KeywordVocabularyURL1.org'
                }
              },
              {
                'keywordValue': {
                  'typeName': 'keywordValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'KeywordTerm2'
                },
                'keywordVocabulary': {
                  'typeName': 'keywordVocabulary',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'KeywordVocabulary2'
                },
                'keywordVocabularyURI': {
                  'typeName': 'keywordVocabularyURI',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://KeywordVocabularyURL2.org'
                }
              }
            ]
          },
          {
            'typeName': 'topicClassification',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'topicClassValue': {
                  'typeName': 'topicClassValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'Topic Classification Term1'
                },
                'topicClassVocab': {
                  'typeName': 'topicClassVocab',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'Topic Classification Vocab1'
                },
                'topicClassVocabURI': {
                  'typeName': 'topicClassVocabURI',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'https://TopicClassificationURL1.com'
                }
              },
              {
                'topicClassValue': {
                  'typeName': 'topicClassValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'Topic Classification Term2'
                },
                'topicClassVocab': {
                  'typeName': 'topicClassVocab',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'Topic Classification Vocab2'
                },
                'topicClassVocabURI': {
                  'typeName': 'topicClassVocabURI',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'https://TopicClassificationURL2.com'
                }
              }
            ]
          },
          {
            'typeName': 'publication',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'publicationCitation': {
                  'typeName': 'publicationCitation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'RelatedPublicationCitation1'
                },
                'publicationIDType': {
                  'typeName': 'publicationIDType',
                  'multiple': False,
                  'typeClass': 'controlledVocabulary',
                  'value': 'ark'
                },
                'publicationIDNumber': {
                  'typeName': 'publicationIDNumber',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'RelatedPublicationIDNumber1'
                },
                'publicationURL': {
                  'typeName': 'publicationURL',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://RelatedPublicationURL1.org'
                }
              },
              {
                'publicationCitation': {
                  'typeName': 'publicationCitation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'RelatedPublicationCitation2'
                },
                'publicationIDType': {
                  'typeName': 'publicationIDType',
                  'multiple': False,
                  'typeClass': 'controlledVocabulary',
                  'value': 'arXiv'
                },
                'publicationIDNumber': {
                  'typeName': 'publicationIDNumber',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'RelatedPublicationIDNumber2'
                },
                'publicationURL': {
                  'typeName': 'publicationURL',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://RelatedPublicationURL2.org'
                }
              }
            ]
          },
          {
            'typeName': 'notesText',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'Notes1'
          },
          {
            'typeName': 'language',
            'multiple': True,
            'typeClass': 'controlledVocabulary',
            'value': [
              'Abkhaz',
              'Afar'
            ]
          },
          {
            'typeName': 'producer',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'producerName': {
                  'typeName': 'producerName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'LastProducer1, FirstProducer1'
                },
                'producerAffiliation': {
                  'typeName': 'producerAffiliation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'ProducerAffiliation1'
                },
                'producerAbbreviation': {
                  'typeName': 'producerAbbreviation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'ProducerAbbreviation1'
                },
                'producerURL': {
                  'typeName': 'producerURL',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://ProducerURL1.org'
                },
                'producerLogoURL': {
                  'typeName': 'producerLogoURL',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://ProducerLogoURL1.org'
                }
              },
              {
                'producerName': {
                  'typeName': 'producerName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'LastProducer2, FirstProducer2'
                },
                'producerAffiliation': {
                  'typeName': 'producerAffiliation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'ProducerAffiliation2'
                },
                'producerAbbreviation': {
                  'typeName': 'producerAbbreviation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'ProducerAbbreviation2'
                },
                'producerURL': {
                  'typeName': 'producerURL',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://ProducerURL2.org'
                },
                'producerLogoURL': {
                  'typeName': 'producerLogoURL',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://ProducerLogoURL2.org'
                }
              }
            ]
          },
          {
            'typeName': 'productionDate',
            'multiple': False,
            'typeClass': 'primitive',
            'value': '1003-01-01'
          },
          {
            'typeName': 'productionPlace',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'ProductionPlace'
            ]
          },
          {
            'typeName': 'contributor',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'contributorType': {
                  'typeName': 'contributorType',
                  'multiple': False,
                  'typeClass': 'controlledVocabulary',
                  'value': 'Data Collector'
                },
                'contributorName': {
                  'typeName': 'contributorName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'LastContributor1, FirstContributor1'
                }
              },
              {
                'contributorType': {
                  'typeName': 'contributorType',
                  'multiple': False,
                  'typeClass': 'controlledVocabulary',
                  'value': 'Data Curator'
                },
                'contributorName': {
                  'typeName': 'contributorName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'LastContributor2, FirstContributor2'
                }
              }
            ]
          },
          {
            'typeName': 'grantNumber',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'grantNumberAgency': {
                  'typeName': 'grantNumberAgency',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'GrantInformationGrantAgency1'
                },
                'grantNumberValue': {
                  'typeName': 'grantNumberValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'GrantInformationGrantNumber1'
                }
              },
              {
                'grantNumberAgency': {
                  'typeName': 'grantNumberAgency',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'GrantInformationGrantAgency2'
                },
                'grantNumberValue': {
                  'typeName': 'grantNumberValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'GrantInformationGrantNumber2'
                }
              }
            ]
          },
          {
            'typeName': 'distributor',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'distributorName': {
                  'typeName': 'distributorName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'LastDistributor1, FirstDistributor1'
                },
                'distributorAffiliation': {
                  'typeName': 'distributorAffiliation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'DistributorAffiliation1'
                },
                'distributorAbbreviation': {
                  'typeName': 'distributorAbbreviation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'DistributorAbbreviation1'
                },
                'distributorURL': {
                  'typeName': 'distributorURL',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://DistributorURL1.org'
                },
                'distributorLogoURL': {
                  'typeName': 'distributorLogoURL',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://DistributorLogoURL1.org'
                }
              },
              {
                'distributorName': {
                  'typeName': 'distributorName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'LastDistributor2, FirstDistributor2'
                },
                'distributorAffiliation': {
                  'typeName': 'distributorAffiliation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'DistributorAffiliation2'
                },
                'distributorAbbreviation': {
                  'typeName': 'distributorAbbreviation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'DistributorAbbreviation2'
                },
                'distributorURL': {
                  'typeName': 'distributorURL',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://DistributorURL2.org'
                },
                'distributorLogoURL': {
                  'typeName': 'distributorLogoURL',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'http://DistributorLogoURL2.org'
                }
              }
            ]
          },
          {
            'typeName': 'distributionDate',
            'multiple': False,
            'typeClass': 'primitive',
            'value': '1004-01-01'
          },
          {
            'typeName': 'depositor',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'LastDepositor, FirstDepositor'
          },
          {
            'typeName': 'dateOfDeposit',
            'multiple': False,
            'typeClass': 'primitive',
            'value': '1002-01-01'
          },
          {
            'typeName': 'timePeriodCovered',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'timePeriodCoveredStart': {
                  'typeName': 'timePeriodCoveredStart',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1005-01-01'
                },
                'timePeriodCoveredEnd': {
                  'typeName': 'timePeriodCoveredEnd',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1005-01-02'
                }
              },
              {
                'timePeriodCoveredStart': {
                  'typeName': 'timePeriodCoveredStart',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1005-02-01'
                },
                'timePeriodCoveredEnd': {
                  'typeName': 'timePeriodCoveredEnd',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1005-02-02'
                }
              }
            ]
          },
          {
            'typeName': 'dateOfCollection',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'dateOfCollectionStart': {
                  'typeName': 'dateOfCollectionStart',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1006-01-01'
                },
                'dateOfCollectionEnd': {
                  'typeName': 'dateOfCollectionEnd',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1006-01-01'
                }
              },
              {
                'dateOfCollectionStart': {
                  'typeName': 'dateOfCollectionStart',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1006-02-01'
                },
                'dateOfCollectionEnd': {
                  'typeName': 'dateOfCollectionEnd',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1006-02-02'
                }
              }
            ]
          },
          {
            'typeName': 'kindOfData',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'KindOfData1',
              'KindOfData2'
            ]
          },
          {
            'typeName': 'series',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'seriesName': {
                  'typeName': 'seriesName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'SeriesName'
                },
                'seriesInformation': {
                  'typeName': 'seriesInformation',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'SeriesInformation'
                }
              }
            ]
          },
          {
            'typeName': 'software',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'softwareName': {
                  'typeName': 'softwareName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'SoftwareName1'
                },
                'softwareVersion': {
                  'typeName': 'softwareVersion',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'SoftwareVersion1'
                }
              },
              {
                'softwareName': {
                  'typeName': 'softwareName',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'SoftwareName2'
                },
                'softwareVersion': {
                  'typeName': 'softwareVersion',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'SoftwareVersion2'
                }
              }
            ]
          },
          {
            'typeName': 'relatedMaterial',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'RelatedMaterial1',
              'RelatedMaterial2'
            ]
          },
          {
            'typeName': 'relatedDatasets',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'RelatedDatasets1',
              'RelatedDatasets2'
            ]
          },
          {
            'typeName': 'otherReferences',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'OtherReferences1',
              'OtherReferences2'
            ]
          },
          {
            'typeName': 'dataSources',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'DataSources1',
              'DataSources2'
            ]
          },
          {
            'typeName': 'originOfSources',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'OriginOfSources'
          },
          {
            'typeName': 'characteristicOfSources',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'CharacteristicOfSourcesNoted'
          },
          {
            'typeName': 'accessToSources',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'DocumentationAndAccessToSources'
          }
        ]
      },
      'geospatial': {
        'displayName': 'Geospatial Metadata',
        'fields': [
          {
            'typeName': 'geographicCoverage',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'country': {
                  'typeName': 'country',
                  'multiple': False,
                  'typeClass': 'controlledVocabulary',
                  'value': 'Afghanistan'
                },
                'state': {
                  'typeName': 'state',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'GeographicCoverageStateProvince1'
                },
                'city': {
                  'typeName': 'city',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'GeographicCoverageCity1'
                },
                'otherGeographicCoverage': {
                  'typeName': 'otherGeographicCoverage',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'GeographicCoverageOther1'
                }
              },
              {
                'country': {
                  'typeName': 'country',
                  'multiple': False,
                  'typeClass': 'controlledVocabulary',
                  'value': 'Albania'
                },
                'state': {
                  'typeName': 'state',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'GeographicCoverageStateProvince2'
                },
                'city': {
                  'typeName': 'city',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'GeographicCoverageCity2'
                },
                'otherGeographicCoverage': {
                  'typeName': 'otherGeographicCoverage',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'GeographicCoverageOther2'
                }
              }
            ]
          },
          {
            'typeName': 'geographicUnit',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'GeographicUnit1',
              'GeographicUnit2'
            ]
          },
          {
            'typeName': 'geographicBoundingBox',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'westLongitude': {
                  'typeName': 'westLongitude',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '-72'
                },
                'eastLongitude': {
                  'typeName': 'eastLongitude',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '-70'
                },
                'northLongitude': {
                  'typeName': 'northLongitude',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '43'
                },
                'southLongitude': {
                  'typeName': 'southLongitude',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '42'
                }
              },
              {
                'westLongitude': {
                  'typeName': 'westLongitude',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '-18'
                },
                'eastLongitude': {
                  'typeName': 'eastLongitude',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '-13'
                },
                'northLongitude': {
                  'typeName': 'northLongitude',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '29'
                },
                'southLongitude': {
                  'typeName': 'southLongitude',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '28'
                }
              }
            ]
          }
        ]
      },
      'socialscience': {
        'displayName': 'Social Science and Humanities Metadata',
        'fields': [
          {
            'typeName': 'unitOfAnalysis',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'UnitOfAnalysis1',
              'UnitOfAnalysis2'
            ]
          },
          {
            'typeName': 'universe',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'Universe1',
              'Universe2'
            ]
          },
          {
            'typeName': 'timeMethod',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'TimeMethod'
          },
          {
            'typeName': 'dataCollector',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'LastDataCollector1, FirstDataCollector1'
          },
          {
            'typeName': 'collectorTraining',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'CollectorTraining'
          },
          {
            'typeName': 'frequencyOfDataCollection',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'Frequency'
          },
          {
            'typeName': 'samplingProcedure',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'SamplingProcedure'
          },
          {
            'typeName': 'targetSampleSize',
            'multiple': False,
            'typeClass': 'compound',
            'value': {
              'targetSampleActualSize': {
                'typeName': 'targetSampleActualSize',
                'multiple': False,
                'typeClass': 'primitive',
                'value': '100'
              },
              'targetSampleSizeFormula': {
                'typeName': 'targetSampleSizeFormula',
                'multiple': False,
                'typeClass': 'primitive',
                'value': 'TargetSampleSizeFormula'
              }
            }
          },
          {
            'typeName': 'deviationsFromSampleDesign',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'MajorDeviationsForSampleDesign'
          },
          {
            'typeName': 'collectionMode',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'CollectionMode'
            ]
          },
          {
            'typeName': 'researchInstrument',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'TypeOfResearchInstrument'
          },
          {
            'typeName': 'dataCollectionSituation',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'CharacteristicsOfDataCollectionSituation'
          },
          {
            'typeName': 'actionsToMinimizeLoss',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'ActionsToMinimizeLosses'
          },
          {
            'typeName': 'controlOperations',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'ControlOperations'
          },
          {
            'typeName': 'weighting',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'Weighting'
          },
          {
            'typeName': 'cleaningOperations',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'CleaningOperations'
          },
          {
            'typeName': 'datasetLevelErrorNotes',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'StudyLevelErrorNotes'
          },
          {
            'typeName': 'responseRate',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'ResponseRate'
          },
          {
            'typeName': 'samplingErrorEstimates',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'EstimatesOfSamplingError'
          },
          {
            'typeName': 'otherDataAppraisal',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'OtherFormsOfDataAppraisal'
          },
          {
            'typeName': 'socialScienceNotes',
            'multiple': False,
            'typeClass': 'compound',
            'value': {
              'socialScienceNotesType': {
                'typeName': 'socialScienceNotesType',
                'multiple': False,
                'typeClass': 'primitive',
                'value': 'NotesType'
              },
              'socialScienceNotesSubject': {
                'typeName': 'socialScienceNotesSubject',
                'multiple': False,
                'typeClass': 'primitive',
                'value': 'NotesSubject'
              },
              'socialScienceNotesText': {
                'typeName': 'socialScienceNotesText',
                'multiple': False,
                'typeClass': 'primitive',
                'value': 'NotesText'
              }
            }
          }
        ]
      },
      'astrophysics': {
        'displayName': 'Astronomy and Astrophysics Metadata',
        'fields': [
          {
            'typeName': 'astroType',
            'multiple': True,
            'typeClass': 'controlledVocabulary',
            'value': [
              'Image',
              'Mosaic',
              'EventList',
              'Cube'
            ]
          },
          {
            'typeName': 'astroFacility',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'Facility1',
              'Facility2'
            ]
          },
          {
            'typeName': 'astroInstrument',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'Instrument1',
              'Instrument2'
            ]
          },
          {
            'typeName': 'astroObject',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'Object1',
              'Object2'
            ]
          },
          {
            'typeName': 'resolution.Spatial',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'SpatialResolution'
          },
          {
            'typeName': 'resolution.Spectral',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'SpectralResolution'
          },
          {
            'typeName': 'resolution.Temporal',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'TimeResolution'
          },
          {
            'typeName': 'coverage.Spectral.Bandpass',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'Bandpass1',
              'Bandpass2'
            ]
          },
          {
            'typeName': 'coverage.Spectral.CentralWavelength',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              '3001',
              '3002'
            ]
          },
          {
            'typeName': 'coverage.Spectral.Wavelength',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'coverage.Spectral.MinimumWavelength': {
                  'typeName': 'coverage.Spectral.MinimumWavelength',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '4001'
                },
                'coverage.Spectral.MaximumWavelength': {
                  'typeName': 'coverage.Spectral.MaximumWavelength',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '4002'
                }
              },
              {
                'coverage.Spectral.MinimumWavelength': {
                  'typeName': 'coverage.Spectral.MinimumWavelength',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '4003'
                },
                'coverage.Spectral.MaximumWavelength': {
                  'typeName': 'coverage.Spectral.MaximumWavelength',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '4004'
                }
              }
            ]
          },
          {
            'typeName': 'coverage.Temporal',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'coverage.Temporal.StartTime': {
                  'typeName': 'coverage.Temporal.StartTime',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1007-01-01'
                },
                'coverage.Temporal.StopTime': {
                  'typeName': 'coverage.Temporal.StopTime',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1007-01-02'
                }
              },
              {
                'coverage.Temporal.StartTime': {
                  'typeName': 'coverage.Temporal.StartTime',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1007-02-01'
                },
                'coverage.Temporal.StopTime': {
                  'typeName': 'coverage.Temporal.StopTime',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1007-02-02'
                }
              }
            ]
          },
          {
            'typeName': 'coverage.Spatial',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'SkyCoverage1',
              'SkyCoverage2'
            ]
          },
          {
            'typeName': 'coverage.Depth',
            'multiple': False,
            'typeClass': 'primitive',
            'value': '200'
          },
          {
            'typeName': 'coverage.ObjectDensity',
            'multiple': False,
            'typeClass': 'primitive',
            'value': '300'
          },
          {
            'typeName': 'coverage.ObjectCount',
            'multiple': False,
            'typeClass': 'primitive',
            'value': '400'
          },
          {
            'typeName': 'coverage.SkyFraction',
            'multiple': False,
            'typeClass': 'primitive',
            'value': '500'
          },
          {
            'typeName': 'coverage.Polarization',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'Polarization'
          },
          {
            'typeName': 'redshiftType',
            'multiple': False,
            'typeClass': 'primitive',
            'value': 'RedshiftType'
          },
          {
            'typeName': 'resolution.Redshift',
            'multiple': False,
            'typeClass': 'primitive',
            'value': '600'
          },
          {
            'typeName': 'coverage.RedshiftValue',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'coverage.Redshift.MinimumValue': {
                  'typeName': 'coverage.Redshift.MinimumValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '701'
                },
                'coverage.Redshift.MaximumValue': {
                  'typeName': 'coverage.Redshift.MaximumValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '702'
                }
              },
              {
                'coverage.Redshift.MinimumValue': {
                  'typeName': 'coverage.Redshift.MinimumValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '703'
                },
                'coverage.Redshift.MaximumValue': {
                  'typeName': 'coverage.Redshift.MaximumValue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '704'
                }
              }
            ]
          }
        ]
      },
      'biomedical': {
        'displayName': 'Life Sciences Metadata',
        'fields': [
          {
            'typeName': 'studyDesignType',
            'multiple': True,
            'typeClass': 'controlledVocabulary',
            'value': [
              'Case Control',
              'Cross Sectional',
              'Cohort Study',
              'Not Specified'
            ]
          },
          {
            'typeName': 'studyFactorType',
            'multiple': True,
            'typeClass': 'controlledVocabulary',
            'value': [
              'Age',
              'Biomarkers',
              'Cell Surface Markers',
              'Developmental Stage'
            ]
          },
          {
            'typeName': 'studyAssayOrganism',
            'multiple': True,
            'typeClass': 'controlledVocabulary',
            'value': [
              'Arabidopsis thaliana',
              'Bos taurus',
              'Caenorhabditis elegans',
              'Danio rerio (zebrafish)'
            ]
          },
          {
            'typeName': 'studyAssayOtherOrganism',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'OtherOrganism1',
              'OtherOrganism2'
            ]
          },
          {
            'typeName': 'studyAssayMeasurementType',
            'multiple': True,
            'typeClass': 'controlledVocabulary',
            'value': [
              'genome sequencing',
              'cell sorting',
              'clinical chemistry analysis',
              'DNA methylation profiling'
            ]
          },
          {
            'typeName': 'studyAssayOtherMeasurmentType',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'OtherMeasurementType1',
              'OtherMeasurementType2'
            ]
          },
          {
            'typeName': 'studyAssayTechnologyType',
            'multiple': True,
            'typeClass': 'controlledVocabulary',
            'value': [
              'culture based drug susceptibility testing, single concentration',
              'culture based drug susceptibility testing, two concentrations',
              'culture based drug susceptibility testing, three or more concentrations (minimum inhibitory concentration measurement)',
              'flow cytometry'
            ]
          },
          {
            'typeName': 'studyAssayPlatform',
            'multiple': True,
            'typeClass': 'controlledVocabulary',
            'value': [
              '210-MS GC Ion Trap (Variant)',
              '220-MS GC Ion Trap (Variant)',
              '225-MS GC Ion Trap (Variant)',
              '300-MS quadrupole GC/MS (Variant)'
            ]
          },
          {
            'typeName': 'studyAssayCellType',
            'multiple': True,
            'typeClass': 'primitive',
            'value': [
              'CellType1',
              'CellType2'
            ]
          }
        ]
      },
      'journal': {
        'displayName': 'Journal Metadata',
        'fields': [
          {
            'typeName': 'journalVolumeIssue',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
              {
                'journalVolume': {
                  'typeName': 'journalVolume',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'JournalVolume1'
                },
                'journalIssue': {
                  'typeName': 'journalIssue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'JournalIssue1'
                },
                'journalPubDate': {
                  'typeName': 'journalPubDate',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1008-01-01'
                }
              },
              {
                'journalVolume': {
                  'typeName': 'journalVolume',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'JournalVolume2'
                },
                'journalIssue': {
                  'typeName': 'journalIssue',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': 'JournalIssue2'
                },
                'journalPubDate': {
                  'typeName': 'journalPubDate',
                  'multiple': False,
                  'typeClass': 'primitive',
                  'value': '1008-02-01'
                }
              }
            ]
          },
          {
            'typeName': 'journalArticleType',
            'multiple': False,
            'typeClass': 'controlledVocabulary',
            'value': 'abstract'
          }
        ]
      }
    }
  }
}



#!/usr/bin/python3
import logging, warnings, os, shutil, tempfile
from pathlib import Path
from urllib import request
from pasta_eln.backend import Backend
from pasta_eln.GUI.project import Project
from pasta_eln.guiCommunicate import Communicate
from pasta_eln.GUI.palette import Palette
from pasta_eln.inputOutput import importELN

def test_simple(qtbot):
  """
  main function
  """
  # initialization: create database, destroy on filesystem and database and then create new one
  warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
  warnings.filterwarnings('ignore', message='invalid escape sequence')
  warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
  warnings.filterwarnings('ignore', category=ImportWarning)
  logPath = Path.home()/'pastaELN.log'
  logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                      datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
  for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
    logging.getLogger(package).setLevel(logging.WARNING)
  logging.info('Start 03 test')

  # start app and load project
  backend = Backend('research')
  palette = Palette(None, 'light_blue')
  comm = Communicate(backend, palette)
  window = Project(comm)
  qtbot.addWidget(window)

  # remove old and recreate empty
  dirName = backend.basePath
  backend.exit()
  try:
    shutil.rmtree(dirName)
    os.makedirs(dirName)
  except Exception:
    pass
  backend = Backend('research')
  tempDir = tempfile.gettempdir()

  allELNs = {'Pasta.eln'          :'PASTA/PASTA.eln',
             'SampleDB.eln'       :'SampleDB/sampledb_export.eln',
             'kadi4mat_1.eln'     :'kadi4mat/collections-example.eln',
             'kadi4mat_2.eln'     :'kadi4mat/records-example.eln',
             'openSemanticLab.eln':'OpenSemanticLab/MinimalExample.osl.eln'
            }
  urlBase = 'https://github.com/SteffenBrinckmann/TheELNFileFormat/raw/refs/heads/new_PastaELN/examples/'
  # 'https://github.com/TheELNConsortium/TheELNFileFormat/raw/refs/heads/master/examples/'
  localDir= '/home/xyz/TheELNConsortium/TheELNFileFormat/examples/'
  local = False
  for eln, pathName in allELNs.items():
    if local:
      elnFileName = f'{localDir}/{pathName}'
    else:
      elnFileName = f'{tempDir}/{eln}'
      request.urlretrieve(f'{urlBase}{pathName}', elnFileName)
    print(f'\nStart with {eln}')
    projName = f'{eln[:-4]} Import'
    backend.addData('x0', {'name': projName})
    df = backend.db.getView('viewDocType/x0')
    projID = df[df['name']==projName]['id'].values[0]
    reply, statistics = importELN(backend, elnFileName, projID)
    print(reply,'\n',statistics)
    assert statistics['num. files']<=statistics['types'].get('File',0) ,'Files do not make sense'
    #statistics type File can be larger than num.Files because there are remote files which are not Files but in the @type
    print(backend.checkDB(minimal=True))

  return



""" Terminology Lookup service """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: terminology_lookup_service.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor
from functools import reduce
from json import load
from os import getcwd
from os.path import dirname, join, realpath
from typing import Any
import requests
from PySide6.QtGui import QPixmap


class TerminologyLookupService:
  """
  Terminology Lookup service which allows user to query for a search term online
  A list of online portals (terminology_lookup_config.json) are queried for the search term
  and the results (dict(information, iri)) are returned
  """

  def __init__(self) -> None:
    self.session_timeout = 10  # Timeout in seconds for the requests send to the lookup services
    current_path = realpath(join(getcwd(), dirname(__file__)))
    self.lookupConfig = join(current_path, 'terminology_lookup_config.json')


  async def do_lookup(self, search_term: str) -> list[dict[str, Any]]:
    """
    Do the lookup for the search term using the services in the terminology_lookup_config.json
    Args:
      search_term (str): Search term used for the lookup

    Returns: List of IRI information for the search term crawled online using the services in the terminology_lookup_config.json
    In case of error, an empty list is returned and errors is updated with failures
    """
    if not search_term.strip():
      return []
    errors = []
    results = []
    # Load the lookup configuration
    with open(self.lookupConfig, encoding='utf-8') as config_file:
      lookup_services = load(config_file)

    # Define a function to perform the HTTP request
    def fetch_service(service: dict[str, Any]) -> dict[str, str] | None:
      try:
        service['request_params'][service['search_term_key']] = search_term
        response = requests.get(service['url'], params=service['request_params'],
          timeout=self.session_timeout)
        if response.status_code == 200:
          web_result = response.json()
          return self.parse_web_result(search_term, web_result, service)
        else:
          errors.append(f"Error querying {service['name']}: {response.status_code} {response.reason}")
      except requests.RequestException as e:
        errors.append(f"Error querying {service['name']}: {str(e)}")
      return None

    # Use ThreadPoolExecutor to run requests in parallel
    loop = asyncio.get_event_loop()
    with ThreadPoolExecutor() as executor:
      tasks = [loop.run_in_executor(executor, fetch_service, service) for service in lookup_services]
      fetched_results = await asyncio.gather(*tasks)
    # Collect valid results
    for result in fetched_results:
      if result:
        results.append(result)
    # Log any errors
    if errors:
      logging.error('Session request errors: %s', errors)
    return results


  def parse_web_result(self, search_term: str, web_result: dict[str, Any], lookup_service: dict[str, Any]) -> dict[str, Any]:
    """
    Parse the web result returned by querying the lookup service for the search term
    Args:
      search_term (str): Search term used for the lookup
      web_result (dict[str, Any]): Web result returned by querying the lookup service
      lookup_service (dict[str, Any]): Lookup service taken from the terminology_lookup_config.json

    Returns (dict[str, Any]): Dictionary containing the name, search term and results which is a list of dict(iri, information)
    """
    if (not search_term or not web_result or not lookup_service):
      logging.error('Invalid search term or web result or lookup service!')
      return {}
    retrieved_results: dict[str, Any] = {'name': lookup_service['name'], 'search_term': search_term, 'results': []}
    # Get mandatory attributes
    result_keys = lookup_service['search_criteria']['results_keys']
    desc_keys = lookup_service['search_criteria']['description_keys']
    id_key = lookup_service['search_criteria']['id_key']
    # Get non mandatory attributes
    duplicate_ontology_names = lookup_service.get('duplicate_ontology_names')
    skip_desc = lookup_service.get('skip_description')
    duplicate_ontology_key = lookup_service['search_criteria'].get('ontology_name_key')
    results = reduce(lambda d, key: d.get(key) if d else None, result_keys,  web_result) # type: ignore[arg-type, return-value]
    for item in results or []:
      description = reduce(lambda d, key: d.get(key) if d else '', desc_keys, item)  # type: ignore[attr-defined]
      is_duplicate = (item[duplicate_ontology_key] in duplicate_ontology_names) if duplicate_ontology_key else False # type: ignore[operator]
      if (description and description != skip_desc and not is_duplicate):
        retrieved_results['results'].append({
            'iri': f"{lookup_service['iri_prefix']}{item[id_key]}"
            if lookup_service.get('iri_prefix') else item[id_key],
            'information': ','.join(description) if isinstance(description, list)
            else description})
    return retrieved_results


  def getIconDict(self) -> dict[str, QPixmap]:
    """
    Get the icon names for the lookup services
    Returns (list[str]): List of icon names for the lookup services
    """
    current_path = realpath(join(getcwd(), dirname(__file__)))
    resources_path = join(current_path, '../../Resources/Icons')
    iconsPixMap = {}
    with open(self.lookupConfig, encoding='utf-8') as config_file:
      for data in load(config_file):
        iconsPixMap[data['name']] = QPixmap(join(resources_path, f'{data["icon_name"]}')).scaledToWidth(50)
    return iconsPixMap



""" Color palette allows easy color access """
import json
import platform
from pathlib import Path
from PySide6.QtWidgets import QApplication, QMainWindow
from qt_material import apply_stylesheet, get_theme  # of https://github.com/UN-GCPDS/qt-material
from ..fixedStringsJson import CONF_FILE_NAME, DEFAULT_COLORS_PALETTE


class Palette():
  """ Color palette allows easy color access """
  def __init__(self, mainWindow:QMainWindow|None, accent:str) -> None:
    """ Initialize the color palette
    Args:
      mainWindow (QMainWindow): main window for getting system theme
      accent (str): accent color, e.g. 'pink'
    """
    accent = self.cleanAccent(accent)                # given theme
    systemTheme = 'light' if mainWindow is None or mainWindow.palette().button().color().red()>128 or \
      platform.system() != 'Linux' else 'dark' # system color mode: dark/light
    self.theme = 'none' if accent=='none' else f'{systemTheme}_{accent}' # theme name
    if accent=='none':
      self.primary       = '#222222'
      self.secondaryText = '#000000'
    else:
      themeData = get_theme(f'{self.theme}.xml')
      self.primary = themeData['primaryColor']
      self.secondaryText = themeData['primaryTextColor']
    # for all themes
    colors = DEFAULT_COLORS_PALETTE[systemTheme]
    self.text = colors['text']
    self.leafX = colors['leafX']
    self.leafO = colors['leafO']
    self.leafShadow = colors['leafShadow']


  def setTheme(self, application:QApplication) -> None:
    """ set theme of application
    Args:
      application (QApplication): application to set the theme
    """
    if self.theme != 'none':
      apply_stylesheet(application, theme=f'{self.theme}.xml')
    return


  def cleanAccent(self, accent:str) -> str:
    """ Clean accent color name

    Args:
      accent (str): accent color, e.g. 'pink'

    Returns:
      str: cleaned accent color
    """
    if '_' in accent:
      newAccent = accent.split('_')[1]
      with open(Path.home()/CONF_FILE_NAME, encoding='utf-8') as fConf:
        configuration =json.load(fConf)
      configuration['GUI']['theme'] = newAccent
      with open(Path.home()/CONF_FILE_NAME, 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(configuration, indent=2))
      return newAccent
    return accent


  def get(self, color: str, prefix: str) -> str:
    """
    Get color with a prefix for CSS styling.

    Args:
        color (str): qt-material colors without the trailing color; 'buttonText' is an additional color.
        prefix (str): CSS key, e.g., 'background-color'.

    Returns:
        str: CSS string, e.g., 'background-color: #333421;'. Returns an empty string if the theme is 'none'.
    """
    if self.theme=='none':
      return ''
    if color=='buttonText':
      return  f'{prefix}: {self.text}; '
    return f'{prefix}: {get_theme(f"{self.theme}.xml")[f"{color}Color"]}; '



""" widget that shows the table of the items """
import itertools
from enum import Enum
from pathlib import Path
from typing import Any
import pandas as pd
import qtawesome as qta
from PySide6.QtWidgets import (QDialog, QFileDialog, QTableWidget,  # pylint: disable=no-name-in-module
                               QTableWidgetItem, QVBoxLayout)
from ...guiCommunicate import Communicate
from ...guiStyle import TextButton, showMessage, space, widgetAndLayout
from ...miscTools import callAddOn
from .key_delegate import KeyDelegate
from .link_online_delegate import LinkOnlineDelegate
from .lookup_delegate import LookupDelegate

COLUMN_NAMES = ['key','long','PURL','', '']
COLUMN_WIDTH = [200,  400,   250, 50, 50]


class Editor(QDialog):
  """ widget that shows the table of the items """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
    """
    super().__init__()
    self.comm = comm
    self.data:pd.DataFrame = pd.DataFrame()
    self.setMinimumWidth(1000)
    self.setMinimumHeight(1000)
    self.setWindowTitle('Edit definitions')

    ### GUI elements
    mainL = QVBoxLayout()
    mainL.setSpacing(space['l'])
    self.setLayout(mainL)
    ### Table
    self.table = QTableWidget(1, 5)
    self.table.verticalHeader().hide()
    self.table.setAlternatingRowColors(True)
    self.table.setHorizontalHeaderLabels(COLUMN_NAMES)
    for idx, width in enumerate(COLUMN_WIDTH):
      self.table.setColumnWidth(idx, width)
    self.keyDelegate = KeyDelegate()
    self.table.setItemDelegateForColumn(0, self.keyDelegate)
    self.linkOnlineDelegate = LinkOnlineDelegate()
    self.table.setItemDelegateForColumn(3, self.linkOnlineDelegate)
    self.lookupDelegate     = LookupDelegate()
    self.table.setItemDelegateForColumn(4, self.lookupDelegate)
    self.table.horizontalHeader().setStretchLastSection(True)
    mainL.addWidget(self.table)
    ### final button box
    _, buttonLineL = widgetAndLayout('H', mainL, 'm')
    TextButton('Import', self, [Command.Import], buttonLineL, 'Import from Excel')
    TextButton('Export', self, [Command.Export], buttonLineL, 'Export to Excel')
    buttonLineL.addStretch(1)
    projectGroup = self.comm.backend.configuration['projectGroups'][self.comm.backend.configurationProjectGroup]
    if 'definition' in projectGroup.get('addOns',{}) and projectGroup['addOns']['definition']:
      TextButton('Autofill PURL',  self, [Command.AddOn], buttonLineL, 'Autofill by using addon')
      buttonLineL.addStretch(1)
    saveBtn = TextButton('Save', self, [Command.Save], buttonLineL, 'Save changes')
    saveBtn.setShortcut('Ctrl+Return')
    TextButton('Cancel', self, [Command.Cancel],   buttonLineL, 'Discard changes')
    ### Data
    db = self.comm.backend.db
    df0 = pd.read_sql_query('SELECT docType, PURL, title FROM docTypes', db.connection).fillna('')
    df0['defType'] = 'class'
    df0 = df0.rename({'docType':'key', 'title':'long'}, axis=1)
    df1 = pd.read_sql_query('SELECT * FROM definitions', db.connection).fillna('')
    df1['defType'] = 'attribute'
    self.data = pd.concat([df0,df1])[['key','long','PURL','defType']]
    self.showDataframe()


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.Export:
      fileName = QFileDialog.getSaveFileName(self, 'Save table to .csv file', str(Path.home()), '*.csv')[0]
      if fileName != '':
        self.getDataframe().to_csv(fileName, index=False)
    elif command[0] is Command.Import:
      fileName = '/home/steffen/test.csv' # QFileDialog.getOpenFileName(self, 'Read table from .csv file', str(Path.home()), '*.csv')[0]
      if fileName != '':
        self.data = pd.read_csv(fileName).fillna('')
        self.showDataframe()
    elif command[0] is Command.AddOn:
      try:
        self.data = callAddOn('definition_autofill', self.comm.backend, self.data, self)
        self.showDataframe()
      except Exception:
        pass
    elif command[0] is Command.Cancel:
      self.reject()
    elif command[0] is Command.Save:
      df = self.getDataframe()
      db = self.comm.backend.db
      for _, row in df.iterrows():
        key, description, purl, dType = row.values
        if dType == 'class':
          try:
            db.cursor.execute(f"UPDATE docTypes SET PURL='{purl}', title='{description}' WHERE docType = '{key}'")
          except Exception:
            showMessage(self, 'ERROR', 'You cannot add a class/docType via this form.', 'Warning')
        else:
          db.cursor.execute('INSERT OR REPLACE INTO definitions VALUES (?, ?, ?);', [key, description, purl])
      db.connection.commit()
      self.accept()
    else:
      print('**ERROR table menu unknown:',command)
    return


  def showDataframe(self) -> None:
    """ Show data frame in the GUI """
    self.table.setRowCount(len(self.data))
    nRows, nCols = self.data.shape
    for i, j in itertools.product(range(nRows), range(nCols-1)):
      rowType = self.data.iloc[i, 3]
      icon = qta.icon('msc.symbol-class' if rowType=='class' else 'msc.symbol-property', scale_factor=1)
      item = QTableWidgetItem(self.data.iloc[i, j])
      if j==0:
        item.setIcon(icon)
      self.table.setItem(i, j, item)
    return


  def getDataframe(self) -> pd.DataFrame:
    """ Get dataframe from table """
    model = self.table.model()
    data = []
    for row in range(model.rowCount()):
      rowRes = [model.index(row, column).data() for column in range(3)]
      data.append(rowRes)
    df = pd.DataFrame(data)
    df = df.rename({0:'key',1:'description',2:'PURL'}, axis=1)
    df = df.merge(self.data, how='left', left_on='key', right_on='key')
    df = df.drop(['long','PURL_y'], axis=1).rename({'PURL_x':'PURL'}, axis=1)
    # df['defType']=self.data['defType'].values    # ignore
    return df


class Command(Enum):
  """ Commands used in this file """
  Save   = 1
  Cancel = 2
  Import = 3
  Export = 4
  AddOn  = 5



""" Custom tree view on data model """
import logging
import os
import platform
import shutil
import subprocess
from enum import Enum
from pathlib import Path
from typing import Any
from PySide6.QtCore import Qt  # pylint: disable=no-name-in-module
from PySide6.QtGui import (QDropEvent, QEventPoint, QStandardItem,  # pylint: disable=no-name-in-module
                           QStandardItemModel)
from PySide6.QtWidgets import (QAbstractItemView, QMenu, QMessageBox, QTreeView,  # pylint: disable=no-name-in-module
                               QWidget)
from ..guiCommunicate import Communicate
from ..guiStyle import Action, showMessage
from ..miscTools import callAddOn
from .projectLeafRenderer import ProjectLeafRenderer


class TreeView(QTreeView):
  """ Custom tree view on data model """
  def __init__(self, parent:QWidget, comm:Communicate, model:QStandardItemModel):
    super().__init__(parent)
    self.aParentWidget = parent
    self.comm = comm
    self.setModel(model)
    self.setHeaderHidden(True)
    self.setStyleSheet('QTreeView::branch {border-image: none;}')
    self.setIndentation(40)
    self.renderer = ProjectLeafRenderer(self.comm)
    self.setItemDelegate(self.renderer)
    self.setExpandsOnDoubleClick(False)
    self.setAcceptDrops(True)
    self.setDropIndicatorShown(True)
    self.setDefaultDropAction(Qt.DropAction.MoveAction)
    self.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
    self.doubleClicked.connect(self.tree2Clicked)


  def contextMenuEvent(self, p:QEventPoint) -> None:                                                         # type: ignore[override]
    """
    create context menu

    Args:
      p (QPoint): point of clicking
    """
    item = self.model().itemFromIndex(self.currentIndex())                                                   # type: ignore[attr-defined]
    if item is None:  #clicked outside any leaf
      return
    folder = item.data()['hierStack'].split('/')[-1][0]=='x'
    context = QMenu(self)
    if folder:
      Action('Add child folder',                   self, [Command.ADD_CHILD],      context)
    Action('Add sibling folder',                   self, [Command.ADD_SIBLING],    context)
    Action('Delete item',                          self, [Command.DELETE],         context)
    context.addSeparator()
    Action('Hide/show item details',               self, [Command.SHOW_DETAILS], context)
    Action('Mark item as hidden/shown',            self, [Command.HIDE],           context)
    context.addSeparator()
    if not folder:
      Action('Open file with another application', self, [Command.OPEN_EXTERNAL],    context)
    Action('Open folder in file browser',          self, [Command.OPEN_FILEBROWSER], context)
    if folder:
      projectGroup = self.comm.backend.configuration['projectGroups'][self.comm.backend.configurationProjectGroup]
      if projectAddOns := projectGroup.get('addOns',{}).get('project',''):
        context.addSeparator()
        for label, description in projectAddOns.items():
          Action(description, self, [Command.ADD_ON, label], context)
    context.exec(p.globalPos())                                                                              # type: ignore[attr-defined]
    return


  def execute(self, command:list[Any]) -> None:
    """
    after selecting a item from context menu

    Args:
      command (list): list of commands
    """
    item = self.model().itemFromIndex(self.currentIndex())                                                   # type: ignore[attr-defined]
    hierStack = item.data()['hierStack'].split('/')
    if command[0] is Command.ADD_CHILD:
      docType= 'x1'
      docID = hierStack[-1]
      self.comm.backend.cwd = Path(self.comm.backend.db.getDoc(docID)['branch'][0]['path'])
      docID = self.comm.backend.addData(docType, {'name':'new item'}, hierStack)['id']
      # append item to the GUI
      item  = self.model().itemFromIndex(self.currentIndex())                                                # type: ignore[attr-defined]
      child = QStandardItem('/'.join(hierStack+[docID]))
      child.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled) # type: ignore
      item.appendRow(child)
      self.comm.changeProject.emit('','') #refresh project
      #appendRow is not 100% correct:
      # - better: insertRow before the first non-folder, depending on the child number
      #   -> get highest non 9999 childNumber
      # turns out, one can easily move it to correct position with drag&drop
      # -  not sure this will be important
    elif command[0] is Command.ADD_SIBLING:
      hierStack= hierStack[:-1]
      docType= 'x1'
      docID  = hierStack[-1]
      self.comm.backend.cwd = Path(self.comm.backend.db.getDoc(docID)['branch'][0]['path'])
      docID = self.comm.backend.addData(docType, {'name':'new item'}, hierStack)['id']
      # append item to the GUI
      item  = self.model().itemFromIndex(self.currentIndex())                                                # type: ignore[attr-defined]
      parent = item.parent() if item.parent() is not None else self.model().invisibleRootItem()              # type: ignore[attr-defined]
      child = QStandardItem('/'.join(hierStack+[docID]))
      child.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled) # type: ignore
      parent.appendRow(child)
      self.comm.changeProject.emit('','') #refresh project
      #++ TODO appendRow is not 100% correct: see above
    elif command[0] is Command.DELETE:
      ret = QMessageBox.critical(self, 'Warning', 'Are you sure you want to delete this data?',\
                                 QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)
      if ret==QMessageBox.StandardButton.Yes:
        docID = hierStack[-1]
        doc = self.comm.backend.db.remove(docID)
        for branch in doc['branch']:
          if branch['path'] is not None:
            oldPath = Path(self.comm.backend.basePath)/branch['path']
            if oldPath.exists():
              newFileName = f'trash_{oldPath.name}'
              if (oldPath.parent/newFileName).exists():  #ensure target does not exist
                endText = ' will be deleted after closing this window. Now, you can still save it to some other place on disk'
                showMessage(self, 'Warning', f'Warning! \nThe file/folder {oldPath.parent/newFileName} {endText}')
                if (oldPath.parent/newFileName).is_file():
                  (oldPath.parent/newFileName).unlink()
                elif (oldPath.parent/newFileName).is_dir():
                  shutil.rmtree(oldPath.parent/newFileName)
              oldPath.rename( oldPath.parent/newFileName)
        # go through children
        children = self.comm.backend.db.getView('viewHierarchy/viewHierarchyAll', startKey='/'.join(doc['branch'][0]['stack']+[docID,'']))
        for line in children:
          self.comm.backend.db.remove(line['id'], stack='/'.join(doc['branch'][0]['stack']+[docID,'']))
        # remove leaf from GUI
        item  = self.model().itemFromIndex(self.currentIndex())                                               # type: ignore[attr-defined]
        parent = item.parent()
        if parent is None: #top level
          parent = self.model().invisibleRootItem()                                                           # type: ignore[attr-defined]
          if parent.rowCount()==1:
            self.aParentWidget.btnAddSubfolder.setVisible(True)                                               # type: ignore[attr-defined]
        parent.removeRow(item.row())
    elif command[0] is Command.SHOW_DETAILS:
      gui    = item.data()['gui']
      gui[0] = not gui[0]
      docID  = hierStack[-1]
      def iterate(currentItem:QStandardItem) -> None:
        """ iterate through all branches and leaves and find items matching the docID
        Args:
          currentItem (QStandardItem): item to iterate to its children
        """
        currentIndex = self.model().indexFromItem(currentItem)                                                # type: ignore[attr-defined]
        if currentItem.data() is not None and docID==currentItem.data()['hierStack'].split('/')[-1]:
          currentItem.setData({ **currentItem.data(), **{'gui':gui}})
        for row in range(self.model().rowCount(currentIndex)):
          for column in range(self.model().columnCount(currentIndex)):
            childIndex = self.model().index(row, column, currentIndex)
            iterate(self.model().itemFromIndex(childIndex))                                                   # type: ignore[attr-defined]
      iterate(self.model().invisibleRootItem())                                                               # type: ignore[attr-defined]
      # only one change once the DB
      self.comm.backend.db.setGUI(docID, gui)
    elif command[0] is Command.HIDE:
      logging.debug('hide document %s',hierStack[-1])
      self.comm.backend.db.hideShow(hierStack[-1])
      #TODO: current implementation: you hide one; all others are hidden as well.
      # Talk to GW what is the default expectation; system allows for individual hiding
      #
      # self.comm.changeProject.emit('','') #refresh project
      # after hide, do not hide immediately but wait on next refresh
    elif command[0] is Command.OPEN_EXTERNAL or command[0] is Command.OPEN_FILEBROWSER:
      # depending if non-folder / folder; address different item in hierstack
      docID = hierStack[-2] \
        if command[0] is Command.OPEN_FILEBROWSER and hierStack[-1][0]!='x' \
        else hierStack[-1]
      doc   = self.comm.backend.db.getDoc(docID)
      if doc['branch'][0]['path'] is None:
        showMessage(self, 'ERROR', 'Cannot open file that is only in the database','Warning')
      else:
        path  = Path(self.comm.backend.basePath)/doc['branch'][0]['path']
        if platform.system() == 'Darwin':       # macOS
          subprocess.call(('open', path))
        elif platform.system() == 'Windows':    # Windows
          os.startfile(path) # type: ignore[attr-defined]
        else:                                   # linux variants
          subprocess.call(('xdg-open', path))
    elif command[0] is Command.ADD_ON:
      callAddOn(command[1], self.comm.backend, item.data()['hierStack'], self)
    else:
      print('**ERROR**: unknown context menu', command)
    return


  def scrollToDoc(self, docID:str) -> None:
    """
    Scroll to document with docID

    Args:
      docID (str): document ID
    """
    def iterate(currentItem:QStandardItem) -> QStandardItem | None:
      """ iterate through all branches and leaves and find items matching the docID
      Args:
        currentItem (QStandardItem): item to iterate to its children
      Returns:
        QStandardItem | None: item with docID or None if not found
      """
      currentIndex = self.model().indexFromItem(currentItem)                                                # type: ignore[attr-defined]
      if currentItem.data() is not None and docID==currentItem.data()['hierStack'].split('/')[-1]:
        return currentItem
      for row in range(self.model().rowCount(currentIndex)):
        for column in range(self.model().columnCount(currentIndex)):
          childIndex = self.model().index(row, column, currentIndex)
          found = iterate(self.model().itemFromIndex(childIndex))                                            # type: ignore[attr-defined]
          if found is not None:
            return found
      return None
    item = iterate(self.model().invisibleRootItem())                                                         # type: ignore[attr-defined]
    if item is not None:
      self.scrollTo(item.index(), QAbstractItemView.EnsureVisible)                                           # type: ignore[attr-defined]
    return


  def tree2Clicked(self) -> None:
    """
    after double-click on tree leaf: open form
    - no redraw required since renderer asks automatically for update
    """
    item = self.model().itemFromIndex(self.currentIndex())                                                   # type: ignore[attr-defined]
    docID = item.data()['hierStack'].split('/')[-1]
    doc   = self.comm.backend.db.getDoc(docID)
    self.comm.formDoc.emit(doc)
    docNew = self.comm.backend.db.getDoc(docID)
    item.setText(docNew['name'])
    item.setData(item.data() | {'docType':docNew['type'], 'gui':docNew['gui']})
    item.emitDataChanged()  #force redraw (resizing and repainting) of this item only
    return


  def dragEnterEvent(self, event:QDropEvent) -> None:
    """
    Override default: what happens if you drag an item

    Args:
      event (QMouseEvent): event
    """
    event.acceptProposedAction()
    return


  def dropEvent(self, event:QDropEvent) -> None:
    """
    Override default: what happens at end of drag&drop

    Args:
      event (QDropEvent): event
    """
    if event.mimeData().hasUrls():  #file rometo to pasta
      item = self.model().itemFromIndex(self.indexAt(event.pos()))                                           # type: ignore[attr-defined]
      if item is None or item.data()['docType'][0][0]!='x':
        showMessage(self, 'Error', 'You can drop external files only onto folders.')
        return
      # create a list of all files
      files, folders = [], []
      for url in event.mimeData().urls():
        path = url.toLocalFile()
        if Path(path).is_file():
          files.append(path)
        else:
          files +=   list(Path(path).rglob('*'))                                                             # type: ignore[arg-type]
          folders += [x[0] for x in os.walk(path)]
      if not (folders+[str(i) for i in files]):
        showMessage(self, 'Error', 'The files seem empty.')
        return
      commonBase   = os.path.commonpath(folders+[str(i) for i in files])
      docID = item.data()['hierStack'].split('/')[-1]
      doc = self.comm.backend.db.getDoc(docID)
      targetFolder = Path(self.comm.backend.cwd/doc['branch'][0]['path'])
      # create folders and copy files
      for folder in folders:
        (targetFolder/(Path(folder).relative_to(commonBase))).mkdir(parents=True, exist_ok=True)
      for fileStr in files:
        file = Path(fileStr)
        if file.is_file():
          shutil.copy(file, targetFolder/(file.relative_to(commonBase)))
      # scan
      for _ in range(2):  #scan twice: convert, extract
        self.comm.backend.scanProject(None, docID, targetFolder.relative_to(self.comm.backend.basePath))
      self.comm.changeProject.emit(item.data()['hierStack'].split('/')[0],'')
      showMessage(self, 'Information','Drag & drop is finished')
      event.ignore()
    elif 'application/x-qstandarditemmodeldatalist' in event.mimeData().formats():
      super().dropEvent(event)
    else:
      logging.error('Drop unknown data: %s', event.mimeData().formats())
    return


class Command(Enum):
  """ Commands used in this file """
  ADD_CHILD        = 1
  ADD_SIBLING      = 2
  DELETE           = 3
  SHOW_DETAILS     = 4
  HIDE             = 5
  OPEN_EXTERNAL    = 6
  OPEN_FILEBROWSER = 7
  ADD_ON           = 8



""" NameColumnDelegate module used for the table views """
from PySide6.QtCore import QModelIndex, QPersistentModelIndex
from PySide6.QtGui import QRegularExpressionValidator
from PySide6.QtWidgets import QLineEdit, QStyledItemDelegate, QStyleOptionViewItem, QWidget


class NameColumnDelegate(QStyledItemDelegate):
  """
  Delegate for creating the line edit with lookup icon for the iri column in data hierarchy editor tables
  """
  def createEditor(self,
                   parent: QWidget,
                   option: QStyleOptionViewItem,
                   index:  QModelIndex | QPersistentModelIndex) -> QWidget:
    """
    Creates the line edit

    Args:
      parent (QWidget): Parent table view.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.

    Returns: QLineEdit widget
    """
    if index.data() in ['name','tags','comment']:
      return QWidget(parent)
    res = QLineEdit(parent)
    res.setValidator(QRegularExpressionValidator(r'[a-z][a-z0-9]+'))
    return res


  def destroyEditor(self, editor:QWidget, index:QModelIndex | QPersistentModelIndex) -> None:
    """ destroy the line editor

    Args:
      editor (QWidget): Parent table view
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.
    """
    if index.row()+1==index.model().rowCount():
      index.model().insertRow(index.model().rowCount())
      index.model().layoutChanged.emit()
    super().destroyEditor(editor, index)
    return



"""html2text: Turn HTML into equivalent Markdown-structured text.

Copied from https://github.com/Alir3z4/html2text
- which has a GPL-3.0 license
- simplified into four files, removed the library constraint
- renamed function

#TODO in here are not by SB
"""
import html.entities
import html.parser
import re
import string
import urllib.parse as urlparse
from textwrap import wrap
from typing import Any, Dict, List, Optional, Tuple, Union
from .html2mdConfig import *
from .html2mdElements import AnchorElement, ListElement
from .html2mdUtils import (dumb_css_parser, element_style, escape_md, escape_md_section, google_fixed_width_font,
                           google_has_height, google_list_style, google_text_emphasis, hn, list_numbering_start,
                           pad_tables_in_text, skipwrap, unifiable_n)

__version__ = (2024, 2, 26)


class HTML2Text(html.parser.HTMLParser):
    def __init__(
        self,
        out: Any = None,
        baseurl: str = '',
        bodywidth: int = BODY_WIDTH,
    ) -> None:
        """
        Input parameters:
            out: possible custom replacement for self.outtextf (which
                 appends lines of text).
            baseurl: base URL of the document we process
        """
        super().__init__(convert_charrefs=False)

        # Config options
        self.split_next_td = False
        self.td_count = 0
        self.table_start = False
        self.unicode_snob = UNICODE_SNOB  # covered in cli
        self.escape_snob = ESCAPE_SNOB  # covered in cli
        self.links_each_paragraph = LINKS_EACH_PARAGRAPH
        self.body_width = bodywidth  # covered in cli
        self.skip_internal_links = SKIP_INTERNAL_LINKS  # covered in cli
        self.inline_links = INLINE_LINKS  # covered in cli
        self.protect_links = PROTECT_LINKS  # covered in cli
        self.google_list_indent = GOOGLE_LIST_INDENT  # covered in cli
        self.ignore_links = IGNORE_ANCHORS  # covered in cli
        self.ignore_mailto_links = IGNORE_MAILTO_LINKS  # covered in cli
        self.ignore_images = IGNORE_IMAGES  # covered in cli
        self.images_as_html = IMAGES_AS_HTML  # covered in cli
        self.images_to_alt = IMAGES_TO_ALT  # covered in cli
        self.images_with_size = IMAGES_WITH_SIZE  # covered in cli
        self.ignore_emphasis = IGNORE_EMPHASIS  # covered in cli
        self.bypass_tables = BYPASS_TABLES  # covered in cli
        self.ignore_tables = IGNORE_TABLES  # covered in cli
        self.google_doc = False  # covered in cli
        self.ul_item_mark = '*'  # covered in cli
        self.emphasis_mark = '_'  # covered in cli
        self.strong_mark = '**'
        self.single_line_break = SINGLE_LINE_BREAK  # covered in cli
        self.use_automatic_links = USE_AUTOMATIC_LINKS  # covered in cli
        self.hide_strikethrough = False  # covered in cli
        self.mark_code = MARK_CODE
        self.wrap_list_items = WRAP_LIST_ITEMS  # covered in cli
        self.wrap_links = WRAP_LINKS  # covered in cli
        self.wrap_tables = WRAP_TABLES
        self.pad_tables = PAD_TABLES  # covered in cli
        self.default_image_alt = DEFAULT_IMAGE_ALT  # covered in cli
        self.tag_callback = None
        self.open_quote = OPEN_QUOTE  # covered in cli
        self.close_quote = CLOSE_QUOTE  # covered in cli
        self.include_sup_sub = INCLUDE_SUP_SUB  # covered in cli

        if out is None:
            self.out = self.outtextf
        else:
            self.out = out

        # empty list to store output characters before they are "joined"
        self.outtextlist: list[str] = []

        self.quiet = 0
        self.p_p = 0  # number of newline character to print before next output
        self.outcount = 0
        self.start = True
        self.space = False
        self.a: list[AnchorElement] = []
        self.astack: list[Optional[dict[str, Optional[str]]]] = []
        self.maybe_automatic_link: Optional[str] = None
        self.empty_link = False
        self.absolute_url_matcher = re.compile(r'^[a-zA-Z+]+://')
        self.account = 0
        self.list: list[ListElement] = []
        self.blockquote = 0
        self.pre = False
        self.startpre = False
        self.code = False
        self.quote = False
        self.br_toggle = ''
        self.lastWasNL = False
        self.lastWasList = False
        self.style = 0
        self.style_def: dict[str, dict[str, str]] = {}
        self.tag_stack: list[tuple[str, dict[str, Optional[str]], dict[str, str]]] = []
        self.emphasis = 0
        self.drop_white_space = 0
        self.inheader = False
        # Current abbreviation definition
        self.abbr_title: Optional[str] = None
        # Last inner HTML (for abbr being defined)
        self.abbr_data: Optional[str] = None
        # Stack of abbreviations to write later
        self.abbr_list: dict[str, str] = {}
        self.baseurl = baseurl
        self.stressed = False
        self.preceding_stressed = False
        self.preceding_data = ''
        self.current_tag = ''

        UNIFIABLE['nbsp'] = '&nbsp_place_holder;'

    def feed(self, data: str) -> None:
        if data is None:
            data = ''
        data = data.replace("</' + 'script>", '</ignore>')
        super().feed(data)

    def handle(self, data: str) -> str:
        self.start = True
        self.feed(data)
        self.feed('')
        markdown = self.optwrap(self.finish())
        if self.pad_tables:
            return pad_tables_in_text(markdown)
        else:
            return markdown

    def outtextf(self, s: str) -> None:
        self.outtextlist.append(s)
        if s:
            self.lastWasNL = s[-1] == '\n'

    def finish(self) -> str:
        self.close()

        self.pbr()
        self.o('', force='end')

        outtext = ''.join(self.outtextlist)

        if self.unicode_snob:
            nbsp = html.entities.html5['nbsp;']
        else:
            nbsp = ' '
        outtext = outtext.replace('&nbsp_place_holder;', nbsp)

        # Clear self.outtextlist to avoid memory leak of its content to
        # the next handling.
        self.outtextlist = []

        return outtext

    def handle_charref(self, c: str) -> None:
        self.handle_data(self.charref(c), True)

    def handle_entityref(self, c: str) -> None:
        ref = self.entityref(c)

        # ref may be an empty string (e.g. for &lrm;/&rlm; markers that should
        # not contribute to the final output).
        # self.handle_data cannot handle a zero-length string right after a
        # stressed tag or mid-text within a stressed tag (text get split and
        # self.stressed/self.preceding_stressed gets switched after the first
        # part of that text).
        if ref:
            self.handle_data(ref, True)

    def handle_starttag(self, tag: str, attrs: list[tuple[str, Optional[str]]]) -> None:
        self.handle_tag(tag, dict(attrs), start=True)

    def handle_endtag(self, tag: str) -> None:
        self.handle_tag(tag, {}, start=False)

    def previousIndex(self, attrs: dict[str, Optional[str]]) -> Optional[int]:
        """
        :type attrs: dict

        :returns: The index of certain set of attributes (of a link) in the
        self.a list. If the set of attributes is not found, returns None
        :rtype: int
        """
        if 'href' not in attrs:
            return None

        match = False
        for i, a in enumerate(self.a):
            if 'href' in a.attrs and a.attrs['href'] == attrs['href']:
                if 'title' in a.attrs or 'title' in attrs:
                    if (
                        'title' in a.attrs
                        and 'title' in attrs
                        and a.attrs['title'] == attrs['title']
                    ):
                        match = True
                else:
                    match = True

            if match:
                return i
        return None

    def handle_emphasis(
        self, start: bool, tag_style: dict[str, str], parent_style: dict[str, str]
    ) -> None:
        """
        Handles various text emphases
        """
        tag_emphasis = google_text_emphasis(tag_style)
        parent_emphasis = google_text_emphasis(parent_style)

        # handle Google's text emphasis
        strikethrough = 'line-through' in tag_emphasis and self.hide_strikethrough

        # google and others may mark a font's weight as `bold` or `700`
        bold = False
        for bold_marker in BOLD_TEXT_STYLE_VALUES:
            bold = bold_marker in tag_emphasis and bold_marker not in parent_emphasis
            if bold:
                break

        italic = 'italic' in tag_emphasis and 'italic' not in parent_emphasis
        fixed = (
            google_fixed_width_font(tag_style)
            and not google_fixed_width_font(parent_style)
            and not self.pre
        )

        if start:
            # crossed-out text must be handled before other attributes
            # in order not to output qualifiers unnecessarily
            if bold or italic or fixed:
                self.emphasis += 1
            if strikethrough:
                self.quiet += 1
            if italic:
                self.o(self.emphasis_mark)
                self.drop_white_space += 1
            if bold:
                self.o(self.strong_mark)
                self.drop_white_space += 1
            if fixed:
                self.o('`')
                self.drop_white_space += 1
                self.code = True
        else:
            if bold or italic or fixed:
                # there must not be whitespace before closing emphasis mark
                self.emphasis -= 1
                self.space = False
            if fixed:
                if self.drop_white_space:
                    # empty emphasis, drop it
                    self.drop_white_space -= 1
                else:
                    self.o('`')
                self.code = False
            if bold:
                if self.drop_white_space:
                    # empty emphasis, drop it
                    self.drop_white_space -= 1
                else:
                    self.o(self.strong_mark)
            if italic:
                if self.drop_white_space:
                    # empty emphasis, drop it
                    self.drop_white_space -= 1
                else:
                    self.o(self.emphasis_mark)
            # space is only allowed after *all* emphasis marks
            if (bold or italic) and not self.emphasis:
                self.o(' ')
            if strikethrough:
                self.quiet -= 1

    def handle_tag(
        self, tag: str, attrs: dict[str, Optional[str]], start: bool
    ) -> None:
        self.current_tag = tag

        if self.tag_callback is not None:
            if self.tag_callback(self, tag, attrs, start) is True:
                return

        # first thing inside the anchor tag is another tag
        # that produces some output
        if (
            start
            and self.maybe_automatic_link is not None
            and tag not in ['p', 'div', 'style', 'dl', 'dt']
            and (tag != 'img' or self.ignore_images)
        ):
            self.o('[')
            self.maybe_automatic_link = None
            self.empty_link = False

        if self.google_doc:
            # the attrs parameter is empty for a closing tag. in addition, we
            # need the attributes of the parent nodes in order to get a
            # complete style description for the current element. we assume
            # that google docs export well formed html.
            parent_style: dict[str, str] = {}
            if start:
                if self.tag_stack:
                    parent_style = self.tag_stack[-1][2]
                tag_style = element_style(attrs, self.style_def, parent_style)
                self.tag_stack.append((tag, attrs, tag_style))
            else:
                dummy, attrs, tag_style = (
                    self.tag_stack.pop() if self.tag_stack else (None, {}, {})
                )
                if self.tag_stack:
                    parent_style = self.tag_stack[-1][2]

        if hn(tag):
            # check if nh is inside of an 'a' tag (incorrect but found in the wild)
            if self.astack:
                if start:
                    self.inheader = True
                    # are inside link name, so only add '#' if it can appear before '['
                    if self.outtextlist and self.outtextlist[-1] == '[':
                        self.outtextlist.pop()
                        self.space = False
                        self.o(hn(tag) * '#' + ' ')
                        self.o('[')
                else:
                    self.p_p = 0  # don't break up link name
                    self.inheader = False
                    return  # prevent redundant emphasis marks on headers
            else:
                self.p()
                if start:
                    self.inheader = True
                    self.o(hn(tag) * '#' + ' ')
                else:
                    self.inheader = False
                    return  # prevent redundant emphasis marks on headers

        if tag in ['p', 'div']:
            if self.google_doc:
                if start and google_has_height(tag_style):
                    self.p()
                else:
                    self.soft_br()
            elif self.astack:
                pass
            elif self.split_next_td:
                pass
            else:
                self.p()

        if tag == 'br' and start:
            if self.blockquote > 0:
                self.o('  \n> ')
            else:
                self.o('  \n')

        if tag == 'hr' and start:
            self.p()
            self.o('* * *')
            self.p()

        if tag in ['head', 'style', 'script']:
            if start:
                self.quiet += 1
            else:
                self.quiet -= 1

        if tag == 'style':
            if start:
                self.style += 1
            else:
                self.style -= 1

        if tag in ['body']:
            self.quiet = 0  # sites like 9rules.com never close <head>

        if tag == 'blockquote':
            if start:
                self.p()
                self.o('> ', force=True)
                self.start = True
                self.blockquote += 1
            else:
                self.blockquote -= 1
                self.p()

        if tag in ['em', 'i', 'u'] and not self.ignore_emphasis:
            # Separate with a space if we immediately follow an alphanumeric
            # character, since otherwise Markdown won't render the emphasis
            # marks, and we'll be left with eg 'foo_bar_' visible.
            # (Don't add a space otherwise, though, since there isn't one in the
            # original HTML.)
            if (
                start
                and self.preceding_data
                and self.preceding_data[-1] not in string.whitespace
                and self.preceding_data[-1] not in string.punctuation
            ):
                emphasis = ' ' + self.emphasis_mark
                self.preceding_data += ' '
            else:
                emphasis = self.emphasis_mark

            self.o(emphasis)
            if start:
                self.stressed = True

        if tag in ['strong', 'b'] and not self.ignore_emphasis:
            # Separate with space if we immediately follow an * character, since
            # without it, Markdown won't render the resulting *** correctly.
            # (Don't add a space otherwise, though, since there isn't one in the
            # original HTML.)
            if (
                start
                and self.preceding_data
                # When `self.strong_mark` is set to empty, the next condition
                # will cause IndexError since it's trying to match the data
                # with the first character of the `self.strong_mark`.
                and len(self.strong_mark) > 0
                and self.preceding_data[-1] == self.strong_mark[0]
            ):
                strong = ' ' + self.strong_mark
                self.preceding_data += ' '
            else:
                strong = self.strong_mark

            self.o(strong)
            if start:
                self.stressed = True

        if tag in ['del', 'strike', 's']:
            if start and self.preceding_data and self.preceding_data[-1] == '~':
                strike = ' ~~'
                self.preceding_data += ' '
            else:
                strike = '~~'

            self.o(strike)
            if start:
                self.stressed = True

        if self.google_doc:
            if not self.inheader:
                # handle some font attributes, but leave headers clean
                self.handle_emphasis(start, tag_style, parent_style)

        if tag in ['kbd', 'code', 'tt'] and not self.pre:
            self.o('`')  # TODO: `` `this` ``
            self.code = not self.code

        if tag == 'abbr':
            if start:
                self.abbr_title = None
                self.abbr_data = ''
                if 'title' in attrs:
                    self.abbr_title = attrs['title']
            else:
                if self.abbr_title is not None:
                    assert self.abbr_data is not None
                    self.abbr_list[self.abbr_data] = self.abbr_title
                    self.abbr_title = None
                self.abbr_data = None

        if tag == 'q':
            if not self.quote:
                self.o(self.open_quote)
            else:
                self.o(self.close_quote)
            self.quote = not self.quote

        def link_url(self: HTML2Text, link: str, title: str = '') -> None:
            url = urlparse.urljoin(self.baseurl, link)
            title = f' "{title}"' if title.strip() else ''
            self.o(f']({escape_md(url)}{title})')

        if tag == 'a' and not self.ignore_links:
            if start:
                if (
                    'href' in attrs
                    and attrs['href'] is not None
                    and not (self.skip_internal_links and attrs['href'].startswith('#'))
                    and not (
                        self.ignore_mailto_links and attrs['href'].startswith('mailto:')
                    )
                ):
                    self.astack.append(attrs)
                    self.maybe_automatic_link = attrs['href']
                    self.empty_link = True
                    if self.protect_links:
                        attrs['href'] = '<' + attrs['href'] + '>'
                else:
                    self.astack.append(None)
            else:
                if self.astack:
                    a = self.astack.pop()
                    if self.maybe_automatic_link and not self.empty_link:
                        self.maybe_automatic_link = None
                    elif a:
                        assert a['href'] is not None
                        if self.empty_link:
                            self.o('[')
                            self.empty_link = False
                            self.maybe_automatic_link = None
                        if self.inline_links:
                            self.p_p = 0
                            title = a.get('title') or ''
                            title = escape_md(title)
                            link_url(self, a['href'], title)
                        else:
                            i = self.previousIndex(a)
                            if i is not None:
                                a_props = self.a[i]
                            else:
                                self.account += 1
                                a_props = AnchorElement(a, self.account, self.outcount)
                                self.a.append(a_props)
                            self.o('][' + str(a_props.count) + ']')

        if tag == 'img' and start and not self.ignore_images:
            if 'src' in attrs and attrs['src'] is not None:
                if not self.images_to_alt:
                    attrs['href'] = attrs['src']
                alt = attrs.get('alt') or self.default_image_alt

                # If we have images_with_size, write raw html including width,
                # height, and alt attributes
                if self.images_as_html or (
                    self.images_with_size and ('width' in attrs or 'height' in attrs)
                ):
                    self.o("<img src='" + attrs['src'] + "' ")
                    if 'width' in attrs and attrs['width'] is not None:
                        self.o("width='" + attrs['width'] + "' ")
                    if 'height' in attrs and attrs['height'] is not None:
                        self.o("height='" + attrs['height'] + "' ")
                    if alt:
                        self.o("alt='" + alt + "' ")
                    self.o('/>')
                    return

                # If we have a link to create, output the start
                if self.maybe_automatic_link is not None:
                    href = self.maybe_automatic_link
                    if (
                        self.images_to_alt
                        and escape_md(alt) == href
                        and self.absolute_url_matcher.match(href)
                    ):
                        self.o('<' + escape_md(alt) + '>')
                        self.empty_link = False
                        return
                    else:
                        self.o('[')
                        self.maybe_automatic_link = None
                        self.empty_link = False

                # If we have images_to_alt, we discard the image itself,
                # considering only the alt text.
                if self.images_to_alt:
                    self.o(escape_md(alt))
                else:
                    self.o('![' + escape_md(alt) + ']')
                    if self.inline_links:
                        href = attrs.get('href') or ''
                        self.o(
                            '(' + escape_md(urlparse.urljoin(self.baseurl, href)) + ')'
                        )
                    else:
                        i = self.previousIndex(attrs)
                        if i is not None:
                            a_props = self.a[i]
                        else:
                            self.account += 1
                            a_props = AnchorElement(attrs, self.account, self.outcount)
                            self.a.append(a_props)
                        self.o('[' + str(a_props.count) + ']')

        if tag == 'dl' and start:
            self.p()
        if tag == 'dt' and not start:
            self.pbr()
        if tag == 'dd' and start:
            self.o('    ')
        if tag == 'dd' and not start:
            self.pbr()

        if tag in ['ol', 'ul']:
            # Google Docs create sub lists as top level lists
            if not self.list and not self.lastWasList:
                self.p()
            if start:
                if self.google_doc:
                    list_style = google_list_style(tag_style)
                else:
                    list_style = tag
                numbering_start = list_numbering_start(attrs)
                self.list.append(ListElement(list_style, numbering_start))
            else:
                if self.list:
                    self.list.pop()
                    if not self.google_doc and not self.list:
                        self.o('\n')
            self.lastWasList = True
        else:
            self.lastWasList = False

        if tag == 'li':
            self.pbr()
            if start:
                if self.list:
                    li = self.list[-1]
                else:
                    li = ListElement('ul', 0)
                if self.google_doc:
                    self.o('  ' * self.google_nest_count(tag_style))
                else:
                    # Indent two spaces per list, except use three spaces for an
                    # unordered list inside an ordered list.
                    # https://spec.commonmark.org/0.28/#motivation
                    # TODO: line up <ol><li>s > 9 correctly.
                    parent_list = None
                    for list in self.list:
                        self.o(
                            '   ' if parent_list == 'ol' and list.name == 'ul' else '  '
                        )
                        parent_list = list.name

                if li.name == 'ul':
                    self.o(self.ul_item_mark + ' ')
                elif li.name == 'ol':
                    li.num += 1
                    self.o(str(li.num) + '. ')
                self.start = True

        if tag in ['table', 'tr', 'td', 'th']:
            if self.ignore_tables:
                if tag == 'tr':
                    if start:
                        pass
                    else:
                        self.soft_br()
                else:
                    pass

            elif self.bypass_tables:
                if start:
                    self.soft_br()
                if tag in ['td', 'th']:
                    if start:
                        self.o(f'<{tag}>\n\n')
                    else:
                        self.o(f'\n</{tag}>')
                else:
                    if start:
                        self.o(f'<{tag}>')
                    else:
                        self.o(f'</{tag}>')

            else:
                if tag == 'table':
                    if start:
                        self.table_start = True
                        if self.pad_tables:
                            self.o('<' + TABLE_MARKER_FOR_PAD + '>')
                            self.o('  \n')
                    else:
                        if self.pad_tables:
                            # add break in case the table is empty or its 1 row table
                            self.soft_br()
                            self.o('</' + TABLE_MARKER_FOR_PAD + '>')
                            self.o('  \n')
                if tag in ['td', 'th'] and start:
                    if self.split_next_td:
                        self.o('| ')
                    self.split_next_td = True

                if tag == 'tr' and start:
                    self.td_count = 0
                if tag == 'tr' and not start:
                    self.split_next_td = False
                    self.soft_br()
                if tag == 'tr' and not start and self.table_start:
                    # Underline table header
                    self.o('|'.join(['---'] * self.td_count))
                    self.soft_br()
                    self.table_start = False
                if tag in ['td', 'th'] and start:
                    self.td_count += 1

        if tag == 'pre':
            if start:
                self.startpre = True
                self.pre = True
            else:
                self.pre = False
                if self.mark_code:
                    self.out('\n[/code]')
            self.p()

        if tag in ['sup', 'sub'] and self.include_sup_sub:
            if start:
                self.o(f'<{tag}>')
            else:
                self.o(f'</{tag}>')

    # TODO: Add docstring for these one letter functions
    def pbr(self) -> None:
        'Pretty print has a line break'
        if self.p_p == 0:
            self.p_p = 1

    def p(self) -> None:
        'Set pretty print to 1 or 2 lines'
        self.p_p = 1 if self.single_line_break else 2

    def soft_br(self) -> None:
        'Soft breaks'
        self.pbr()
        self.br_toggle = '  '

    def o(
        self, data: str, puredata: bool = False, force: Union[bool, str] = False
    ) -> None:
        """
        Deal with indentation and whitespace
        """
        if self.abbr_data is not None:
            self.abbr_data += data

        if not self.quiet:
            if self.google_doc:
                # prevent white space immediately after 'begin emphasis'
                # marks ('**' and '_')
                lstripped_data = data.lstrip()
                if self.drop_white_space and not (self.pre or self.code):
                    data = lstripped_data
                if lstripped_data != '':
                    self.drop_white_space = 0

            if puredata and not self.pre:
                # This is a very dangerous call ... it could mess up
                # all handling of &nbsp; when not handled properly
                # (see entityref)
                data = re.sub(r'\s+', r' ', data)
                if data and data[0] == ' ':
                    self.space = True
                    data = data[1:]
            if not data and not force:
                return

            if self.startpre:
                # self.out(" :") #TODO: not output when already one there
                if not data.startswith('\n') and not data.startswith('\r\n'):
                    # <pre>stuff...
                    data = '\n' + data
                if self.mark_code:
                    self.out('\n[code]')
                    self.p_p = 0

            bq = '>' * self.blockquote
            if not (force and data and data[0] == '>') and self.blockquote:
                bq += ' '

            if self.pre:
                if not self.list:
                    bq += '    '
                # else: list content is already partially indented
                bq += '    ' * len(self.list)
                data = data.replace('\n', '\n' + bq)

            if self.startpre:
                self.startpre = False
                if self.list:
                    # use existing initial indentation
                    data = data.lstrip('\n')

            if self.start:
                self.space = False
                self.p_p = 0
                self.start = False

            if force == 'end':
                # It's the end.
                self.p_p = 0
                self.out('\n')
                self.space = False

            if self.p_p:
                self.out((self.br_toggle + '\n' + bq) * self.p_p)
                self.space = False
                self.br_toggle = ''

            if self.space:
                if not self.lastWasNL:
                    self.out(' ')
                self.space = False

            if self.a and (
                (self.p_p == 2 and self.links_each_paragraph) or force == 'end'
            ):
                if force == 'end':
                    self.out('\n')

                newa = []
                for link in self.a:
                    if self.outcount > link.outcount:
                        self.out(
                            '   ['
                            + str(link.count)
                            + ']: '
                            + urlparse.urljoin(self.baseurl, link.attrs['href'])
                        )
                        if 'title' in link.attrs and link.attrs['title'] is not None:
                            self.out(' (' + link.attrs['title'] + ')')
                        self.out('\n')
                    else:
                        newa.append(link)

                # Don't need an extra line when nothing was done.
                if self.a != newa:
                    self.out('\n')

                self.a = newa

            if self.abbr_list and force == 'end':
                for abbr, definition in self.abbr_list.items():
                    self.out('  *[' + abbr + ']: ' + definition + '\n')

            self.p_p = 0
            self.out(data)
            self.outcount += 1

    def handle_data(self, data: str, entity_char: bool = False) -> None:
        if not data:
            # Data may be empty for some HTML entities. For example,
            # LEFT-TO-RIGHT MARK.
            return

        if self.stressed:
            data = data.strip()
            self.stressed = False
            self.preceding_stressed = True
        elif self.preceding_stressed:
            if (
                re.match(r'[^][(){}\s.!?]', data[0])
                and not hn(self.current_tag)
                and self.current_tag not in ['a', 'code', 'pre']
            ):
                # should match a letter or common punctuation
                data = ' ' + data
            self.preceding_stressed = False

        if self.style:
            self.style_def.update(dumb_css_parser(data))

        if self.maybe_automatic_link is not None:
            href = self.maybe_automatic_link
            if (
                href == data
                and self.absolute_url_matcher.match(href)
                and self.use_automatic_links
            ):
                self.o('<' + data + '>')
                self.empty_link = False
                return
            else:
                self.o('[')
                self.maybe_automatic_link = None
                self.empty_link = False

        if not self.code and not self.pre and not entity_char:
            data = escape_md_section(data, snob=self.escape_snob)
        self.preceding_data = data
        self.o(data, puredata=True)

    def charref(self, name: str) -> str:
        if name[0] in ['x', 'X']:
            c = int(name[1:], 16)
        else:
            c = int(name)

        if not self.unicode_snob and c in unifiable_n:
            return unifiable_n[c]
        else:
            try:
                return chr(c)
            except ValueError:  # invalid unicode
                return ''

    def entityref(self, c: str) -> str:
        if not self.unicode_snob and c in UNIFIABLE:
            return UNIFIABLE[c]
        try:
            ch = html.entities.html5[c + ';']
        except KeyError:
            return '&' + c + ';'
        return UNIFIABLE[c] if c == 'nbsp' else ch

    def google_nest_count(self, style: dict[str, str]) -> int:
        """
        Calculate the nesting count of google doc lists

        :type style: dict

        :rtype: int
        """
        nest_count = 0
        if 'margin-left' in style:
            nest_count = int(style['margin-left'][:-2]) // self.google_list_indent

        return nest_count

    def optwrap(self, text: str) -> str:
        """
        Wrap all paragraphs in the provided text.

        :type text: str

        :rtype: str
        """
        if not self.body_width:
            return text

        result = ''
        newlines = 0
        # I cannot think of a better solution for now.
        # To avoid the non-wrap behaviour for entire paras
        # because of the presence of a link in it
        if not self.wrap_links:
            self.inline_links = False
        for para in text.split('\n'):
            if len(para) > 0:
                if not skipwrap(
                    para, self.wrap_links, self.wrap_list_items, self.wrap_tables
                ):
                    indent = ''
                    if para.startswith('  ' + self.ul_item_mark):
                        # list item continuation: add a double indent to the
                        # new lines
                        indent = '    '
                    elif para.startswith('> '):
                        # blockquote continuation: add the greater than symbol
                        # to the new lines
                        indent = '> '
                    wrapped = wrap(
                        para,
                        self.body_width,
                        break_long_words=False,
                        subsequent_indent=indent,
                    )
                    result += '\n'.join(wrapped)
                    if para.endswith('  '):
                        result += '  \n'
                        newlines = 1
                    elif indent:
                        result += '\n'
                        newlines = 1
                    else:
                        result += '\n\n'
                        newlines = 2
                else:
                    # Warning for the tempted!!!
                    # Be aware that obvious replacement of this with
                    # line.isspace()
                    # DOES NOT work! Explanations are welcome.
                    if not RE_SPACE.match(para):
                        result += para + '\n'
                        newlines = 1
            else:
                if newlines < 2:
                    result += '\n'
                    newlines += 1
        return result


def html2markdown(html: str, baseurl: str = '', bodywidth: Optional[int] = None) -> str:
    if bodywidth is None:
        bodywidth = BODY_WIDTH
    h = HTML2Text(baseurl=baseurl, bodywidth=bodywidth)
    return h.handle(html).strip()



"""
Minisign
from https://github.com/x13a/py-minisign
     https://jedisct1.github.io/minisign
"""

__version__ = '0.10'

from .exceptions import Error, ParseError, VerifyError
from .minisign import KeyPair, PublicKey, SecretKey, Signature



""" Main class of config tab on GUI elements """
import json
from pathlib import Path
from typing import Callable
from PySide6.QtWidgets import (QDialog, QDialogButtonBox, QFormLayout, QGroupBox,  # pylint: disable=no-name-in-module
                               QVBoxLayout)
from ..fixedStringsJson import CONF_FILE_NAME, configurationGUI
from ..guiCommunicate import Communicate
from ..guiStyle import TextButton, addRowList
from ..miscTools import restart


class ConfigurationGUI(QDialog):
  """ Main class of config tab on GUI elements """
  def __init__(self, comm:Communicate, callbackFinished:Callable[[bool],None]):
    """
    Initialization

    Args:
      comm (Communicate): communication
      callbackFinished (function): callback function to call upon end
    """
    super().__init__()
    self.comm = comm
    self.callbackFinished = callbackFinished
    #GUI elements
    if hasattr(self.comm.backend, 'configuration'):
      onDisk = self.comm.backend.configuration['GUI']
      mainL  = QVBoxLayout(self)
      for label, items  in configurationGUI.items():
        groupbox = QGroupBox(label.capitalize())
        mainL.addWidget(groupbox)
        sectionL = QFormLayout(groupbox)
        for k,v in items.items():
          setattr(self, k, addRowList(sectionL, label=v[0], default=str(onDisk[k]), itemList=[str(i) for i in v[2]]))
          getattr(self, k).setStyleSheet(self.comm.palette.get('secondaryText', 'color'))
    #final button box
    buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel)
    buttonBox.clicked.connect(self.closeDialog)
    mainL.addWidget(buttonBox)


  def closeDialog(self, btn:TextButton) -> None:
    """
    Save changes to hard-disk
    """
    if btn.text().endswith('Cancel'):
      self.reject()
      self.callbackFinished(False)
    else:
      for items in configurationGUI.values():
        for k in items.keys():
          try:
            self.comm.backend.configuration['GUI'][k] = int(getattr(self, k).currentText())
          except Exception:
            self.comm.backend.configuration['GUI'][k] = getattr(self, k).currentText()
      with open(Path.home()/CONF_FILE_NAME, 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.comm.backend.configuration,indent=2))
      restart()
    return



"""extract data from vendor
- javascript object notion

Tutorial:
- make sure the content is not to large: here 0.3MB is the cutoff for the content
- for other tutorials, see extractor_csv.py and extractor_png.py
"""
import json
import os


def use(filePath, style={'main':''}, saveFileName=None):
  """
  Args:
    filePath (Path): full path file name
    style    (dict): supplied to guide the display / extraction style = recipe
                     main is / separated hierarchical elements parent->child
                     can contain more elements
    saveFileName (string): if given, save the image to this file-name

  Returns:
    dict: containing image, metaVendor, metaUser, recipe
  """
  # Extractor for fancy instrument
  content = ''
  fileSize = os.stat(filePath).st_size / (1024*1024)  #in MB
  if fileSize<0.3:
    with open(filePath, encoding='utf-8') as jsonFile:
      jsonContent = jsonFile.read()
      content = json.loads(jsonContent)
      if not isinstance(content, dict):
        content= {'content': json.loads(jsonContent)}
  else:
    content= 'Too large json file'
  style['main'] = 'procedure/json'
  content= f'```json\n{json.dumps(content, indent=2)}\n```'

  # return everything
  return {'content':content, 'style':style, 'metaVendor':{}, 'metaUser':{}}
